/*******************************************************************************
*  MCom-02 FreeRTOS port                                                       *
*  Copyright (c) 2017 by RnD Center "ELVEES", JSC                              *
*                                                                              *
*  This file is part of the FreeRTOS distribution.                             *
*                                                                              *
*  FreeRTOS is free software; you can redistribute it and/or modify it under   *
*  the terms of the GNU General Public License (version 2) as published by the *
*  Free Software Foundation >>>> AND MODIFIED BY <<<< the FreeRTOS exception.  *
*******************************************************************************/


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* Scheduler include files. */
#include "freertos/source/FreeRTOS.h"
#include "freertos/source/task.h"
#include "freertos/source/semphr.h"
#include "freertos/source/queue.h"

/* FreeRTOS+TCP includes. */
#include "freertos/source/FreeRTOS_TCP/FreeRTOS_IP.h"
#include "freertos/source/FreeRTOS_TCP/FreeRTOS_Sockets.h"
#include "freertos/source/FreeRTOS_TCP/FreeRTOS_DHCP.h"
#include "freertos/source/FreeRTOS_Plus_TCP_Demos/include/SimpleTCPEchoServer.h"
#include "freertos/source/FreeRTOS_Plus_TCP_Demos/include/TCPEchoClient_SingleTasks.h"
#include "freertos/source/FreeRTOS_Plus_CLI_Demos/include/TCPCommandConsole.h"
#include "freertos/source/FreeRTOS_Plus_CLI_Demos/include/UDPCommandConsole.h"
#include "freertos/source/FreeRTOS-Plus-CLI/FreeRTOS_CLI.h"

/* Other includes */
#include "platform.h"
#include "drivers/dbgled/dbgled.h"
#include "drivers/spi/spi_flash.h"
#include "services/uartsrv.h"
#include "services/netsrv.h"
#include "services/sdsrv.h"
#include "freertos/source/FreeRTOS_FAT/ff_stdio.h"
#include "tasks/examples.h"

/* Enable interactive mode */
#define USE_INTERACTIVE_MODE    1

/* Enable modules */
#define TCP_EN                  1
#define SD_EN                   1

#if( TESTS_ENABLE == 1 )
#include "tests/fat/test_fat.h"
#include "tests/sd/test_sd.h"
#include "tests/memory/test_memory.h"
#include "tests/float_point/float_point_test.h"

static BaseType_t prvTestSD( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );
static BaseType_t prvTestFATRAM( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );
static BaseType_t prvTestFATSD( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );
static BaseType_t prvTestMemory( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );
static BaseType_t prvTestUserTask( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );
static BaseType_t prvTestFloatPoint( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString );

static const CLI_Command_Definition_t xTestSDEnableCmd = 
{
        "test-sd",
        "\r\ntest-sd:\r\n Run SD card write-read-compare test\r\n",
        prvTestSD,
        0
};

static const CLI_Command_Definition_t xTestFATRAMEnableCmd = 
{
        "test-fat-ram",
        "\r\ntest-fat-ram:\r\n Run FAT test on RAM disk\r\n",
        prvTestFATRAM,
        0
};

static const CLI_Command_Definition_t xTestFATSDEnableCmd = 
{
        "test-fat-sd",
        "\r\ntest-fat-sd:\r\n Run FAT test on SD disk\r\n",
        prvTestFATSD,
        0
};

static const CLI_Command_Definition_t xTestMemory = 
{
        "test-memory",
        "\r\ntest-memory:\r\n Run memory test\r\n",
        prvTestMemory,
        0
};

static const CLI_Command_Definition_t xTestFloatPoint =
{
    "test-float-point",
    "\r\ntest-float-point:\r\n Run float point test. This test use superPI function.\r\n",
    prvTestFloatPoint, 
    0 
};

#if( configUSE_USER_MODE_SUPPORT == 1 )
static const CLI_Command_Definition_t xTestUserTask = 
{
        "test-user-task",
        "\r\ntest-user-task <asid>:\r\n Run user task test with corresponding asid\r\n",
        prvTestUserTask,
        1
};

extern int _user_entry;
extern int __user_start__;
extern int __user_end__;
#endif /* configUSE_USER_MODE_SUPPORT == 1 */
#endif /* TESTS_ENABLE == 1 */

/* Prototypes */
#if(SD_EN == 1)
extern void vRegisterSDSrvCLICommands( void );
#endif
#if(TCP_EN == 1)
extern void vRegisterNetsrvCLICommands( void );
#endif
extern void vRegisterUserTasksCommands( void );
void initServices( void );

/* 
 * Prototypes for the standard FreeRTOS callback/hook functions implemented
 * within this file. 
 */
void vApplicationMallocFailedHook( void );
void vApplicationIdleHook( void );
void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName );
void vApplicationTickHook( void );

/* Variables */
static UBaseType_t ulNextRand;  /* Use by the pseudo random number generator. */
volatile void* xSemaphorePrintf = 0; 
extern volatile void* rxStoreSemaphore;

int main(int argc, char** argv) 
{
BaseType_t xResult;
    (void)xResult; /* Remove compiler warning about unused parameter */
      
    initPlatform();
       
    if(_get_cpu_id() == 0) 
    {
        initServices();
    }
    else 
    {
        /* Entry point for CPU1 */
        while(1)
        {
            volatile int i;
            for(i = 0; i < 10000000; i++) asm volatile( "nop");
            switchDBGLED(0);
        }
    }
    vInitialiseRunTimeStats();
    
    #if( TESTS_ENABLE == 1)
        #if( SD_EN == 1 )
            FreeRTOS_CLIRegisterCommand( &xTestFATSDEnableCmd );
            FreeRTOS_CLIRegisterCommand( &xTestSDEnableCmd );
        #endif
        #if( configUSE_USER_MODE_SUPPORT == 1 )
            FreeRTOS_CLIRegisterCommand( &xTestUserTask );
        #endif
        FreeRTOS_CLIRegisterCommand( &xTestFATRAMEnableCmd );
        FreeRTOS_CLIRegisterCommand( &xTestMemory );
        FreeRTOS_CLIRegisterCommand( &xTestFloatPoint );
    #endif
    
    /*
     * IF USE_INTERACTIVE_MODE == 1 system register CLI service run function.
     * Services run if you call it explicit using CLI.
     * ELSE run services immediately.
     */
    #if( USE_INTERACTIVE_MODE == 1 )
        #if( TCP_EN == 1 )
            vRegisterNetsrvCLICommands();
        #endif
        #if( SD_EN == 1 )
            vRegisterSDSrvCLICommands();
        #endif
        vRegisterSampleCLICommands();
        vRegisterUserTasksCommands();
    #else
        #if( TCP_EN == 1 )
            vNetsrvStartTCPServer();
        #endif
    #endif

    vUARTCommandConsoleStart( netsrvUDP_CLI_TASK_STACK_SIZE, netsrvUDP_CLI_TASK_PRIORITY );
    vTaskStartScheduler();
    
    for( ;; );
    
    return 0;
}
/*-----------------------------------------------------------*/

void initServices( void )
{
    /* Init printf semaphore */
    _api_uart_config(ROM_BAUD115200);
    rxStoreSemaphore = xSemaphoreCreateBinary();
    configASSERT(rxStoreSemaphore != NULL);
    xSemaphoreGive( rxStoreSemaphore );  
    xSemaphorePrintf = xSemaphoreCreateBinary();
    configASSERT(xSemaphorePrintf != NULL);
    xSemaphoreGive( xSemaphorePrintf );  
    vInitUARTService(configUART_CONSOLE_OUTPUT_NUM, UART_SRV_REALTIME_POLICY);
    #if( ( SD_EN == 1 ) && ( USE_INTERACTIVE_MODE == 0 )  )
        vInitSDService( 0 );
    #endif    
}
/*-----------------------------------------------------------*/

void vApplicationMallocFailedHook( void )
{
	/* Called if a call to pvPortMalloc() fails because there is insufficient
	free memory available in the FreeRTOS heap.  pvPortMalloc() is called
	internally by FreeRTOS API functions that create tasks, queues, software
	timers, and semaphores.  The size of the FreeRTOS heap is set by the
	configTOTAL_HEAP_SIZE configuration constant in FreeRTOSConfig.h. */
        uart_putstr(configUART_CONSOLE_OUTPUT_NUM, "Error :: Malloc Failed!\r\n\0");
	taskDISABLE_INTERRUPTS();
	for( ;; );
}
/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName )
{
	( void ) pcTaskName;
	( void ) pxTask;
        char buffer[128];
	/* Run time stack overflow checking is performed if
	configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
	function is called if a stack overflow is detected. */
        sprintf(buffer, "Error :: StackOverFlow :: %s\r\n", pcTaskName);
        uart_putstr(configUART_CONSOLE_OUTPUT_NUM, buffer);
	taskDISABLE_INTERRUPTS();
	for( ;; );
}
/*-----------------------------------------------------------*/

void vApplicationIdleHook( void )
{

}
/*-----------------------------------------------------------*/

void vAssertCalled( const char * pcFile, unsigned long ulLine )
{
    char buffer[128];
    
	( void ) pcFile;
	( void ) ulLine;

	taskENTER_CRITICAL();
	{
            sprintf(buffer, "  ******* ASSERT :: %s : %d *******\r\n", pcFile, (int)ulLine);
            uart_putstr(configUART_CONSOLE_OUTPUT_NUM, buffer);
	}
        buffer[0] = 0;
        while(1);
	taskEXIT_CRITICAL();
}
/*-----------------------------------------------------------*/

void vApplicationTickHook( void )
{
	
}

/*-----------------------------------------------------------*/

void vApplicationGetIdleTaskMemory( StaticTask_t **ppxIdleTaskTCBBuffer, StackType_t **ppxIdleTaskStackBuffer, uint16_t *pusIdleTaskStackSize )
{
	/* configUSE_STATIC_ALLOCATION is set to 1, so the application has the
	opportunity to supply the buffers that will be used by the Idle task as its
	stack and to hold its TCB.  If these are set to NULL then the buffers will
	be allocated dynamically, just as if xTaskCreate() had been called. */
	*ppxIdleTaskTCBBuffer = NULL;
	*ppxIdleTaskStackBuffer = NULL;
	*pusIdleTaskStackSize = configMINIMAL_STACK_SIZE; /* In words.  NOT in bytes! */
}
/*-----------------------------------------------------------*/

void vApplicationGetTimerTaskMemory( StaticTask_t **ppxTimerTaskTCBBuffer, StackType_t **ppxTimerTaskStackBuffer, uint16_t *pusTimerTaskStackSize )
{
	/* configUSE_STATIC_ALLOCATION is set to 1, so the application has the
	opportunity to supply the buffers that will be used by the Timer/RTOS daemon
	task as its	stack and to hold its TCB.  If these are set to NULL then the
	buffers will be allocated dynamically, just as if xTaskCreate() had been
	called. */
	*ppxTimerTaskTCBBuffer = NULL;
	*ppxTimerTaskStackBuffer = NULL;
	*pusTimerTaskStackSize = configTIMER_TASK_STACK_DEPTH; /* In words.  NOT in bytes! */
}
/*-----------------------------------------------------------*/

UBaseType_t uxRand( void )
{
const uint32_t ulMultiplier = 0x015a4e35UL, ulIncrement = 1UL;

	/* Utility function to generate a pseudo random number. */
	ulNextRand = ( ulMultiplier * ulNextRand ) + ulIncrement;
	return( ( int ) ( ulNextRand >> 16UL ) & 0x7fffUL );
}
/*-----------------------------------------------------------*/

const char *pcApplicationHostnameHook( void )
{
	/* Assign the name "rtosdemo" to this network node.  This function will be
	called during the DHCP: the machine will be registered with an IP address
	plus this name. */
	return configHOST_NAME;
}
/*-----------------------------------------------------------*/

#if( TESTS_ENABLE == 1 )
static BaseType_t prvTestSD( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{
    ( void ) pcCommandString;
    snprintf( pcWriteBuffer, xWriteBufferLen, lTestSD() ? "test-sd :: fail\r\n" : "test-sd :: success\r\n" );

    return pdFALSE;
}
/*-----------------------------------------------------------*/

static BaseType_t prvTestFATRAM( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{
    ( void ) pcCommandString;
    snprintf( pcWriteBuffer, xWriteBufferLen, lTestFAT( netsrvRAM_DISK_NAME ) ? "test-fat-ram :: fail\r\n" : "test-fat-ram :: success\r\n" );

    return pdFALSE;
}
/*-----------------------------------------------------------*/

static BaseType_t prvTestFATSD( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{
    ( void ) pcCommandString;
    snprintf( pcWriteBuffer, xWriteBufferLen, lTestFAT( sdsrvSD_DISK_NAME ) ? "test-fat-sd :: fail\r\n" : "test-fat-sd :: success\r\n" );

    return pdFALSE;
}
/*-----------------------------------------------------------*/

static BaseType_t prvTestMemory( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{
    ( void ) pcCommandString;
    snprintf( pcWriteBuffer, xWriteBufferLen, lTestCallocFree( testmemoryKBYTES, testmemorySTEP_BYTES ) ? "test-memory :: fail\r\n" : "test-memory :: success\r\n" );

    return pdFALSE;
}
/*-----------------------------------------------------------*/

static BaseType_t prvTestFloatPoint( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{   
char * pcMsg = "float-test task started\n";

    xTaskCreate( vFloatPointTestTask, "float-point", 0x400, NULL, tskIDLE_PRIORITY + 1, NULL );			
 
    xWriteBufferLen = strlen( pcMsg ) + 1;
    snprintf( pcWriteBuffer, xWriteBufferLen, "%s\n", pcMsg );
    
    return pdFALSE;
}
/*-----------------------------------------------------------*/

#if( configUSE_USER_MODE_SUPPORT == 1 )
static BaseType_t prvTestUserTask( char *pcWriteBuffer, size_t xWriteBufferLen, const char *pcCommandString )
{
char * pcMsg = "test-user-task :: started\n";
BaseType_t xParameterLen = 0;
const char *pcParam = NULL;    
BaseType_t xResult = pdFALSE;
uint8_t ucASID = 0;
char cTaskName[ 32 ] = { 0 };
uint32_t ulCodeSize = 0;
void *pvUserSegment = NULL;
    
    ( void ) pcCommandString;
   	
    xWriteBufferLen = strlen( pcMsg ) + 1;
    snprintf( pcWriteBuffer, xWriteBufferLen, "%s\n", pcMsg );
    
    pcParam = FreeRTOS_CLIGetParameter( pcCommandString, 1, &xParameterLen );
    ucASID = atoi( pcParam );
    
    if( ucASID == 0 )
    {
        FreeRTOS_printf(( "test-user-task :: error, invalid ASID\n" ));
    }

    pvUserSegment = alloc_user_segment( ucASID, 1 );
    if( pvUserSegment == NULL )
    {
        FreeRTOS_printf(( "test-user-task :: error, can't allocate user segment\n" ));
        return pdFALSE;
    }
    
    //copy user code
    ulCodeSize = (unsigned int)&__user_end__ - (unsigned int)&__user_start__;
    memcpy(pvUserSegment, &__user_start__, ulCodeSize);
    
    invalidate_icache();

    sprintf(cTaskName, "UserTask.%x", ucASID);
    xResult = xUserModeTaskCreate( ucASID,                      /* ASID of user task */
                                   (void*)0x0,                  /* The function that implements the task. */
                                   (const char*)cTaskName, 	/* The text name assigned to the task - for debug only as it is not used by the kernel. */
                                   configMINIMAL_STACK_SIZE, 	/* The size of the stack to allocate to the task. */
                                   ucASID,              	/* The parameter passed to the task - not used in this case. */
                                   tskIDLE_PRIORITY + 1, 	/* The priority assigned to the task. */
                                   NULL );			/* The task handle is not required, so NULL is passed. */
    
    if( xResult != pdTRUE )
    {
        FreeRTOS_printf(( "test-user-task :: error, can't create user task\n" ));
    }
  
    return pdFALSE;
}
/*-----------------------------------------------------------*/
#endif /* configUSE_USER_MODE_SUPPORT == 1 */
#endif /* TESTS_ENABLE == 1 */

